import typing
from traceback import print_exc

import duckdb
import pyarrow as pa
import sqlglot
from pyarrow import ChunkedArray, Table
from snowflake.connector.constants import FIELD_TYPES, FIELD_NAME_TO_ID
from snowflake.connector.cursor import ResultMetadataV2

from universql.util import SnowflakeError

class DuckDBFunctions:
    @staticmethod
    def register(db: duckdb.DuckDBPyConnection):
        db.create_function("CURRENT_WAREHOUSE", DuckDBFunctions.current_warehouse, [], duckdb.typing.VARCHAR)
        pass

    @staticmethod
    def current_warehouse() -> str:
        return "x-duck"


def get_field_for_snowflake(column: ResultMetadataV2, value: typing.Optional[pa.Array] = None) -> \
        typing.Tuple[
            pa.Field, pa.Array]:
    arrow_field = FIELD_TYPES[column.type_code]

    metadata = {
        "logicalType": arrow_field.name,
        "charLength": "8388608",
        "byteLength": "8388608",
    }

    if arrow_field.name == "GEOGRAPHY":
        metadata["logicalType"] = "OBJECT"

    precision = column.precision
    scale = column.precision

    if arrow_field.name == 'FIXED':
        ### Rmove this line
        if value is not None and pa.types.is_integer(value.type):
            pa_type = value.type
        else:
            ###
            pa_type = pa.decimal128(column.precision
                                    # or value.type.precision
                                    or 38,
                                    column.scale
                                    # or value.type.precision
                                    or 0)
        precision = precision or 1
        scale = scale or 0
        if value is not None:
            try:
                value = value.cast(pa_type, safe=False)
            except Exception as e:
                raise e
    elif arrow_field.name == 'DATE':
        pa_type = pa.date32()
        if value is not None:
            value = value.cast(pa_type)
    elif arrow_field.name == 'TIME':
        pa_type = pa.int64()
        if value is not None:
            value = value.cast(pa_type)
    elif arrow_field.name == 'TIMESTAMP_LTZ' or arrow_field.name == 'TIMESTAMP_NTZ' or arrow_field.name == 'TIMESTAMP':
        metadata["final_type"] = "T"
        timestamp_fields = [
            pa.field("epoch", nullable=False, type=pa.int64(), metadata=metadata),
            pa.field("fraction", nullable=False, type=pa.int32(), metadata=metadata),
        ]
        pa_type = pa.struct(timestamp_fields)
        if value is not None:
            epoch = pa.compute.divide(value.cast(pa.int64()), 1_000_000_000).combine_chunks()
            value = pa.StructArray.from_arrays(arrays=[epoch, pa.nulls(len(value), type=pa.int32())],
                                               fields=timestamp_fields)
    elif arrow_field.name == 'TIMESTAMP_TZ':
        metadata["final_type"] = "T"
        timestamp_fields = [
            pa.field("epoch", nullable=False, type=pa.int64(), metadata=metadata),
            pa.field("fraction", nullable=False, type=pa.int32(), metadata=metadata),
            pa.field("timezone", nullable=False, type=pa.int32(), metadata=metadata),
        ]
        pa_type = pa.struct(timestamp_fields)
        if value is not None:
            epoch = pa.compute.divide(value.cast(pa.int64()), 1_000_000_000).combine_chunks()

            value = pa.StructArray.from_arrays(
                arrays=[epoch,
                        # TODO: modulos 1_000_000_000 to get the fraction of a second, pyarrow doesn't support the operator yet
                        pa.nulls(len(value), type=pa.int32()),
                        # TODO: reverse engineer the timezone conversion
                        pa.nulls(len(value), type=pa.int32()),
                        ],
                fields=timestamp_fields)
    else:
        pa_type = arrow_field.pa_type(column)

    if precision is not None:
        metadata["precision"] = str(precision)
    if scale is not None:
        metadata["scale"] = str(scale)

    field = pa.field(column.name, type=pa_type, nullable=column.is_nullable, metadata=metadata)
    return (field, value)


def arrow_to_snowflake_type_id(column_type: pa.DataType):
    if pa.types.is_decimal(column_type) or pa.types.is_integer(column_type):
        type_code = FIELD_NAME_TO_ID["FIXED"]
    elif pa.types.is_date(column_type):
        type_code = FIELD_NAME_TO_ID["DATE"]
    elif pa.types.is_floating(column_type):
        type_code = FIELD_NAME_TO_ID["REAL"]
    elif pa.types.is_timestamp(column_type):
        # Check if it should be TIMESTAMP_TZ
        type_code = FIELD_NAME_TO_ID["TIMESTAMP"]
    elif pa.types.is_boolean(column_type):
        type_code = FIELD_NAME_TO_ID["BOOLEAN"]
    elif pa.types.is_string(column_type):
        type_code = FIELD_NAME_TO_ID["TEXT"]
    elif pa.types.is_struct(column_type):
        # Check if it should be MAP
        type_code = FIELD_NAME_TO_ID["VARIANT"]
    elif pa.types.is_list(column_type):
        type_code = FIELD_NAME_TO_ID["ARRAY"]
    elif pa.types.is_binary(column_type):
        type_code = FIELD_NAME_TO_ID["BINARY"]
    elif pa.types.is_time(column_type):
        type_code = FIELD_NAME_TO_ID["TIME"]
    else:
        # Unsupported types: VECTOR, GEOMETRY, GEOGRAPHY
        raise ValueError(f"Unsupported type: {column_type}")

    return type_code


def get_field_from_duckdb(column: list[str], arrow_table: Table, idx: int) -> typing.Tuple[
    typing.Optional[ChunkedArray], pa.Field]:
    (field_name, field_type) = column[0], column[1]
    pa_type = arrow_table.schema[idx].type

    metadata = {}
    value = arrow_table[idx]

    if field_type == 'NUMBER':

        if (  # no harm for int types
                pa_type != pa.int64() and
                pa_type != pa.int32() and
                pa_type != pa.int16() and
                pa_type != pa.int8()):
            pa_type = pa.decimal128(getattr(value.type, 'precision', 38), getattr(value.type, 'scale', 0))
        try:
            value = value.cast(pa_type)
        except Exception as e:
            pass
        metadata["logicalType"] = "FIXED"
        metadata["precision"] = "1"
        metadata["scale"] = "0"
        metadata["physicalType"] = "SB1"
        metadata["final_type"] = "T"
    elif field_type == 'Date':
        pa_type = pa.date32()
        value = value.cast(pa_type)
        metadata["logicalType"] = "DATE"
    elif field_type == 'Time':
        pa_type = pa.int64()
        value = value.cast(pa_type)
        metadata["logicalType"] = "TIME"
    elif field_type == "BINARY":
        pa_type = pa.binary()
        metadata["logicalType"] = "BINARY"
    elif field_type == "TIMESTAMP" or field_type == "DATETIME" or field_type == "TIMESTAMP_LTZ":
        metadata["logicalType"] = "TIMESTAMP_LTZ"
        metadata["precision"] = "0"
        metadata["scale"] = "9"
        metadata["physicalType"] = "SB16"
        metadata["final_type"] = "T"
        timestamp_fields = [
            pa.field("epoch", nullable=False, type=pa.int64(), metadata=metadata),
            pa.field("fraction", nullable=False, type=pa.int32(), metadata=metadata),
        ]
        pa_type = pa.struct(timestamp_fields)
        epoch = pa.compute.divide(value.cast(pa.int64()), 1_000_000_000).combine_chunks()
        value = pa.StructArray.from_arrays(arrays=[epoch, pa.nulls(len(value), type=pa.int32())],
                                           fields=timestamp_fields)
    elif field_type == "TIMESTAMP_NTZ":
        metadata["logicalType"] = "TIMESTAMP_NTZ"
        metadata["precision"] = "0"
        metadata["scale"] = "9"
        metadata["physicalType"] = "SB16"
        timestamp_fields = [
            pa.field("epoch", nullable=False, type=pa.int64(), metadata=metadata),
            pa.field("fraction", nullable=False, type=pa.int32(), metadata=metadata),
        ]
        pa_type = pa.struct(timestamp_fields)
        epoch = pa.compute.divide(value.cast(pa.int64()), 1_000_000_000).combine_chunks()
        value = pa.StructArray.from_arrays(arrays=[epoch, pa.nulls(len(value), type=pa.int32())],
                                           fields=timestamp_fields)
    elif field_type == "TIMESTAMP_TZ":
        timestamp_fields = [
            pa.field("epoch", nullable=False, type=pa.int64(), metadata=metadata),
            pa.field("fraction", nullable=False, type=pa.int32(), metadata=metadata),
            pa.field("timezone", nullable=False, type=pa.int32(), metadata=metadata),
        ]
        pa_type = pa.struct(timestamp_fields)
        epoch = pa.compute.divide(value.cast(pa.int64()), 1_000_000_000).combine_chunks()

        value = pa.StructArray.from_arrays(
            arrays=[epoch,
                    # TODO: modulos 1_000_000_000 to get the fraction of a second, pyarrow doesn't support the operator yet
                    pa.nulls(len(value), type=pa.int32()),
                    # TODO: reverse engineer the timezone conversion
                    pa.nulls(len(value), type=pa.int32()),
                    ],
            fields=timestamp_fields)
        metadata["logicalType"] = "TIMESTAMP_TZ"
        metadata["precision"] = "0"
        metadata["scale"] = "9"
        metadata["physicalType"] = "SB16"
    elif field_type == "JSON":
        pa_type = pa.utf8()
        metadata["logicalType"] = "OBJECT"
        metadata["charLength"] = "16777216"
        metadata["byteLength"] = "16777216"
        metadata["scale"] = "0"
        metadata["precision"] = "38"
        metadata["finalType"] = "T"
    elif pa_type == pa.bool_():
        metadata["logicalType"] = "BOOLEAN"
    elif field_type == 'list':
        pa_type = pa.utf8()
        arrow_to_project = duckdb.from_arrow(arrow_table.select([field_name]))
        metadata["logicalType"] = "ARRAY"
        metadata["charLength"] = "16777216"
        metadata["byteLength"] = "16777216"
        metadata["scale"] = "0"
        metadata["precision"] = "38"
        metadata["finalType"] = "T"
        value = (arrow_to_project.project(f"to_json({field_name})").arrow())[0]
    elif pa_type == pa.string():
        metadata["logicalType"] = "TEXT"
        metadata["charLength"] = "16777216"
        metadata["byteLength"] = "16777216"
    else:
        raise Exception()

    field = pa.field(field_name, type=pa_type, nullable=True, metadata=metadata)
    return value, field



import logging
import os
import typing
from typing import List

import duckdb
import snowflake
import sqlglot
from fakesnow.conn import FakeSnowflakeConnection
from fakesnow.cursor import FakeSnowflakeCursor
from pyiceberg.catalog import load_catalog, PY_CATALOG_IMPL
from pyiceberg.exceptions import NoSuchTableError
from pyiceberg.io import PY_IO_IMPL, WAREHOUSE
from snowflake.connector.options import pyarrow
from sqlglot.expressions import Select, Insert, Create, Drop, Properties, TemporaryProperty, Schema, Table

from universql.warehouse import ICatalog, Executor, IcebergTable, Locations, CreateRelation, Location
from universql.lake.cloud import s3, gcs, CACHE_DIRECTORY_KEY
from universql.util import prepend_to_lines, QueryError, calculate_script_cost
from universql.protocol.utils import DuckDBFunctions, get_field_from_duckdb
from sqlglot.optimizer.simplify import simplify

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("🐥")

class DuckDBCatalog(ICatalog):

    def register_locations(self, tables: Locations):
        pass

    def __init__(self, context: dict, query_id: str, credentials: dict, compute: dict):
        super().__init__(context, query_id, credentials, compute)
        duckdb_path = context.get('database_path')
        self.duckdb = duckdb.connect(duckdb_path, config={
            'max_memory': context.get('max_memory'),
            'temp_directory': os.path.join(context.get('cache_directory'), "duckdb-staging"),
            'max_temp_directory_size': context.get('max_cache_size'),
        })
        DuckDBFunctions.register(self.duckdb)
        self.duckdb.install_extension("iceberg")
        self.duckdb.load_extension("iceberg")
        iceberg_catalog = context.get('iceberg_catalog')
        if iceberg_catalog is not None:
            catalog_name = iceberg_catalog
            catalog_props = {}
        else:
            catalog_name = None
            catalog_props = {
                PY_IO_IMPL: "universql.lake.cloud.iceberg",
                WAREHOUSE: "gs://my-iceberg-data/custom-events/customer_iceberg_pyiceberg",
                PY_CATALOG_IMPL: "pyiceberg.catalog.sql.SqlCatalog",
                CACHE_DIRECTORY_KEY: context.get('cache_directory'),
                "uri": "sqlite:////Users/bkabak/Code/universql/tests/test.db",
                # "echo": "true"
            }
        self.iceberg_catalog = load_catalog(catalog_name, **catalog_props)

        fake_snowflake_conn = FakeSnowflakeConnection(self.duckdb, "main", "public", False, False)
        fake_snowflake_conn.database_set = True
        fake_snowflake_conn.schema_set = True
        self.emulator = FakeSnowflakeCursor(fake_snowflake_conn, self.duckdb)
        self._register_data_lake(context)

    def _register_data_lake(self, args: dict):
        self.duckdb.register_filesystem(s3(args.get('cache_directory'), args.get('aws_profile')))
        self.duckdb.register_filesystem(gcs(args.get('cache_directory'), args.get('gcp_project')))

    def get_table_paths(self, tables: List[sqlglot.exp.Table]):
        pass

    def executor(self) -> Executor:
        return DuckDBExecutor(self)


class DuckDBExecutor(Executor):

    def __init__(self, duckdb: DuckDBCatalog):
        self.catalog = duckdb

    def get_query_log(self, total_duration) -> str:
        cost = calculate_script_cost(total_duration)
        return f"Run locally on DuckDB: {cost}"

    def supports(self, ast: sqlglot.exp.Expression) -> bool:
        return isinstance(ast, Select) or isinstance(ast, Insert) or isinstance(ast, Create)

    def execute_raw(self, raw_query: str) -> None:
        try:
            self.catalog.emulator.execute(raw_query)
        except duckdb.Error as e:
            raise QueryError(f"Unable to run the parse locally on DuckDB. {e.args}")
        except duckdb.duckdb.DatabaseError as e:
            raise QueryError(f"Unable to run the query locally on DuckDB. {e.msg}")
        except snowflake.connector.ProgrammingError as e:
            raise QueryError(f"Unable to run the query locally on DuckDB. {e.msg}")

    def _get_iceberg_ref(self, default_namespace, destination_table, ):
        namespace = '.'.join([part.sql() for part in destination_table.parts[0:-1]])
        if namespace == '':
            namespace = default_namespace
            table_ref = f"{default_namespace}.{destination_table.sql()}"
        else:
            namespace = namespace
            table_ref = destination_table.sql()
        return namespace, table_ref

    def _sync_catalog(self, ast: sqlglot.exp.Expression,
                      tables_getter: typing.Callable[[], Locations]) -> sqlglot.exp.Expression:
            locations = tables_getter()
            return self._sync_duckdb_catalog(locations,
                                             simplify(
                                                 ast)) if locations is not None else None

    def execute(self, ast: sqlglot.exp.Expression, tables_getter: typing.Callable[[], Locations]) -> \
            typing.Optional[Locations]:

        if isinstance(ast, Create) or isinstance(ast, Insert):
            destination_table = ast.this
            if not isinstance(destination_table, Table):
                if isinstance(destination_table, Schema):
                    destination_table = destination_table.this
                    # filter columns
                else:
                    raise QueryError(f"CREATE {ast.kind} is not supported in DuckDB")

            (catalog, schema, table) = destination_table.parts
            catalog = sqlglot.exp.parse_identifier(catalog or self.catalog.credentials.get('database')).sql()
            schema = sqlglot.exp.parse_identifier(schema or self.catalog.credentials.get('schema')).sql()

            if isinstance(ast, Create):
                if ast.kind == 'TABLE':
                    properties = destination_table.args.get('properties') or Properties()
                    is_temp = TemporaryProperty() in properties.expressions

                    self.catalog.iceberg_catalog.create_namespace_if_not_exists((catalog, schema))
                    self.execute_raw(self._sync_catalog(ast.expression, tables_getter).sql(dialect="duckdb"))
                    arrow_table = self.get_as_table()
                    create_iceberg_table = self.catalog.iceberg_catalog.create_table_if_not_exists(table.sql(),
                                                                                                   arrow_table.schema)
                    create_iceberg_table.append(arrow_table)

                    if is_temp:
                        self.catalog.duckdb.register(destination_table.sql(), arrow_table)
                    else:
                        return {destination_table: IcebergTable(create_iceberg_table.metadata_location)}
                elif ast.kind == 'VIEW':
                    properties = destination_table.args.get('properties') or Properties()
                    is_temp = TemporaryProperty() in properties.expressions
                    duckdb_query = self._sync_catalog(ast, tables_getter).expression.sql(dialect="duckdb")
                    self.execute_raw(duckdb_query)

                    if not is_temp:
                        return {destination_table: CreateRelation(ast.args.get('properties'), ast.kind, ast.expression)}

            elif isinstance(ast, Insert):
                try:
                    iceberg_table = self.catalog.iceberg_catalog.load_table((catalog, schema, table.sql()))
                except NoSuchTableError as e:
                    raise QueryError(f"Error accessing catalog {e.args}")
                self.execute_raw(self._sync_catalog(ast.expression, tables_getter).sql(dialect="duckdb"))
                table = self.get_as_table()
                iceberg_table.append(table)

        elif isinstance(ast, Drop):
            delete_table = ast.this
            self.catalog.iceberg_catalog.drop_table(delete_table.sql())
        else:
            sql = self._sync_catalog(ast, tables_getter).sql(dialect="duckdb", pretty=True)
            self.execute_raw(sql)

        return None

    def get_as_table(self) -> pyarrow.Table:
        arrow_table = self.catalog.emulator._arrow_table
        if arrow_table is None:
            raise QueryError("No result returned from DuckDB")
        for idx, column in enumerate(self.catalog.duckdb.description):
            array, schema = get_field_from_duckdb(column, arrow_table, idx)
            arrow_table = arrow_table.set_column(idx, schema, array)
        return arrow_table

    def close(self):
        self.catalog.emulator.close()

    def _sync_duckdb_catalog(self, locations: typing.Dict[sqlglot.exp.Table, str], ast: sqlglot.exp.Expression) -> \
            typing.Optional[
                sqlglot.exp.Expression]:

        schemas = set((table.catalog, table.db) for table in locations.keys() if table.db != '')
        databases = set(table[0] for table in schemas if table[0] != '')

        databases_sql = [f"ATTACH IF NOT EXISTS ':memory:' AS {sqlglot.exp.parse_identifier(database).sql()}" for
                         database in
                         databases]
        schemas_sql = [
            f"CREATE SCHEMA IF NOT EXISTS {sqlglot.exp.parse_identifier(db).sql()}.{sqlglot.exp.parse_identifier(schema).sql()}"
            for
            (db, schema) in schemas]
        views_sql = [f"CREATE OR REPLACE VIEW {table.sql()} AS SELECT * FROM {self.get_iceberg_read(location)};" for
                     table, location
                     in locations.items()]
        views_sql = ";\n".join(databases_sql + schemas_sql + views_sql)

        if views_sql != "":
            logger.info(
                f"[{self.catalog.query_id}] DuckDB environment is setting up:\n{prepend_to_lines(views_sql)}")
            try:
                self.catalog.duckdb.execute(views_sql)
            except duckdb.Error as e:
                raise QueryError(f"Unable to sync DuckDB catalog. {e.args}")
                pass

        # def replace_icebergs_with_duckdb_reference(
        #         expression: sqlglot.exp.Expression) -> sqlglot.exp.Expression:
        #     if isinstance(expression, sqlglot.exp.Table):
        #         if expression.name != "":
        #             new_table = sqlglot.exp.to_table(f"main.{sqlglot.exp.parse_identifier(expression.sql())}")
        #             return new_table
        #             # return locations[expression]
        #         else:
        #             return expression
        #
        #     return expression

        return (ast
                # .transform(replace_icebergs_with_duckdb_reference)
                .transform(self.fix_snowflake_to_duckdb_types))

    @staticmethod
    def fix_snowflake_to_duckdb_types(
            expression: sqlglot.exp.Expression) -> sqlglot.exp.Expression:
        if isinstance(expression, sqlglot.exp.DataType):
            if expression.this.value in ["TIMESTAMPLTZ", "TIMESTAMPTZ"]:
                return sqlglot.exp.DataType.build("TIMESTAMPTZ")
            if expression.this.value in ["VARIANT"]:
                return sqlglot.exp.DataType.build("JSON")

        return expression

    @staticmethod
    def get_iceberg_read(location: Location) -> str:
        if isinstance(location, IcebergTable):
            return sqlglot.exp.func("iceberg_scan",
                                    sqlglot.exp.Literal.string(location.location)).sql()

        raise QueryError("Unsupported op")



import logging
import time

import pyarrow
import sqlglot
from sqlglot import ParseError
from sqlglot.expressions import Create

from universql.util import get_friendly_time_since, \
    prepend_to_lines, parse_compute, QueryError
from universql.warehouse import Executor
from universql.warehouse.bigquery import BigQueryCatalog
from universql.warehouse.duckdb import DuckDBCatalog
from universql.warehouse.snowflake import SnowflakeCatalog

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("💡")

COMPUTES = {"duckdb": DuckDBCatalog, "local": DuckDBCatalog, "bigquery": BigQueryCatalog, "snowflake": SnowflakeCatalog}


class UniverSQLSession:
    def __init__(self, context, token, credentials: dict, session_parameters: dict):
        self.context = context
        self.credentials = credentials
        self.session_parameters = [{"name": item[0], "value": item[1]} for item in session_parameters.items()]
        self.token = token
        self.compute_plan = parse_compute(self.credentials.get("warehouse"))
        first_catalog_compute = next(filter(lambda x: x.get("name") == 'snowflake', self.compute_plan), {}).get('args')
        self.catalog = SnowflakeCatalog(context, self.token, self.credentials, first_catalog_compute)
        self.catalog_executor = self.catalog.executor()
        self.computes = {"snowflake": self.catalog_executor}

        self.last_executor_cursor = None
        self.processing = False

    def _must_run_on_catalog(self, tables, ast):
        queries_that_doesnt_need_warehouse = ["show"]
        if ast.name in queries_that_doesnt_need_warehouse \
                or ast.key in queries_that_doesnt_need_warehouse:
            return True
        for table in tables:
            if (len(table.parts) == 1 and self.credentials.get('schema') == "information_schema"
                    or len(table.parts) > 1 and isinstance(table.parts[-2].this, str) and
                    table.parts[-2].this.lower() == 'information_schema'
                    or len(table.parts) > 2 and isinstance(table.parts[-3].this, str)
                    and table.parts[-3].this.lower() == "snowflake"):
                return True
        return False

    def _do_query(self, start_time: float, raw_query: str) -> pyarrow.Table:

        try:
            queries = sqlglot.parse(raw_query, read="snowflake")
        except ParseError as e:
            raise QueryError(f"Unable to parse query with SQLGlot: {e.args}")

        last_executor = None

        if queries is None:
            last_executor = self.perform_query(self.catalog_executor, raw_query)
        else:
            last_error = None
            for ast in queries:
                for compute in self.compute_plan:
                    last_error = None
                    compute_name = compute.get('name')
                    last_executor = self.computes.get(compute_name)
                    if last_executor is None:
                        last_executor = self.computes[compute_name] = COMPUTES[compute_name](self.context,
                                                                                             self.token,
                                                                                             self.credentials,
                                                                                             compute).executor()
                    try:
                        last_executor = self.perform_query(last_executor, raw_query, ast=ast)
                        break
                    except QueryError as e:
                        logger.warning(f"Unable to run query on {compute_name}: {e.message}")
                        last_error = e

            performance_counter = time.perf_counter()
            query_duration = performance_counter - start_time

            if last_error is not None:
                raise last_error

            logger.info(
                f"[{self.token}] {last_executor.get_query_log(query_duration)} 🚀 ({get_friendly_time_since(start_time, performance_counter)})")

        return last_executor.get_as_table()

    def perform_query(self, alternative_executor: Executor, raw_query, ast=None) -> Executor:
        if (ast is not None and alternative_executor.supports(ast) and
                alternative_executor != self.catalog_executor):
            if isinstance(ast, Create):
                tables = list(ast.expression.find_all(sqlglot.exp.Table))
            else:
                tables = list(ast.find_all(sqlglot.exp.Table))

            must_run_on_catalog = self._must_run_on_catalog(tables, ast)
            if not must_run_on_catalog:
                new_locations = alternative_executor.execute(ast, lambda: self.catalog.get_table_paths(tables))
                if new_locations is not None:
                    self.catalog.register_locations(new_locations)
                return alternative_executor

        last_executor = self.catalog_executor
        if ast is None:
            last_executor.execute_raw(raw_query)
        else:
            last_executor.execute(ast, lambda: {})
        return last_executor

    def do_query(self, raw_query: str) -> pyarrow.Table:
        start_time = time.perf_counter()
        logger.info(f"[{self.token}] Transpiling query \n{prepend_to_lines(raw_query)}")
        self.processing = True
        try:
            return self._do_query(start_time, raw_query)
        finally:
            # logger.info(
            # f"[{self.token}] 🌊 Done processing in total {get_friendly_time_since(start_time, time.perf_counter())}")
            self.processing = False

    def close(self):
        self.catalog_executor.close()



import typing
from abc import ABC, abstractmethod
from typing import List

import sqlglot
from snowflake.connector.options import pyarrow
from sqlglot.expressions import Property


class IcebergTable:
    def __init__(self, location):
        self.location = location


class CreateRelation:
    def __init__(self, properties: List[Property], kind: str, query: sqlglot.exp.Expression):
        self.properties = properties
        self.kind = kind
        self.query = query


Location = CreateRelation | IcebergTable | pyarrow.Table
Locations = typing.Dict[sqlglot.exp.Table, Location]


class Executor(ABC):
    @abstractmethod
    def supports(self, ast: sqlglot.exp.Expression) -> bool:
        pass

    @abstractmethod
    def execute(self, ast: sqlglot.exp.Expression, locations: typing.Callable[[], Locations]) -> \
            typing.Optional[
                typing.Dict[sqlglot.exp.Table, str]]:
        pass

    @abstractmethod
    def execute_raw(self, raw_query: str) -> None:
        pass

    @abstractmethod
    def get_as_table(self) -> pyarrow.Table:
        pass

    @abstractmethod
    def get_query_log(self, total_duration) -> str:
        pass

    @abstractmethod
    def close(self):
        pass


class ICatalog(ABC):
    def __init__(self, context, query_id: str, credentials: dict, compute: dict):
        self.context = context
        self.query_id = query_id
        self.credentials = credentials
        self.compute = compute

    def create(self, ast: sqlglot.exp.Table, table: pyarrow.Table):
        pass

    @abstractmethod
    def get_table_paths(self, tables: List[sqlglot.exp.Table]) -> typing.Dict[sqlglot.exp.Table, str]:
        pass

    @abstractmethod
    def register_locations(self, tables: Locations):
        pass

    @abstractmethod
    def executor(self) -> Executor:
        pass



import streamlit as st

st.title("Stlite Sharing: Serverless Streamlit app platform")

st.markdown("""
### Stlite
**Stlite** is a port of _Streamlit_ to Wasm, powered by Pyodide,
that runs completely on web browsers.

The official repository is [🔗 here](https://github.com/whitphx/stlite).

If you are new to Streamlit, read the Getting Started tutorial [🔗 here](https://docs.streamlit.io/library/get-started) first
(don't worry, it only takes a few minutes 👍),
but **you can skip the "Installation" section** because you are here 😎.
You can start writing code right out of the box on this online editor 👈!
(If there is not an editor on the left, you are seeing the shared app.
Navigate to the editor mode: https://edit.share.stlite.net/)

### Stlite Sharing
This page is built on **Stlite Sharing**, an online code editor & sharing platform for _Stlite_. \\
If you see the editor and preview panes side by side, you are in the editor mode, https://edit.share.stlite.net/. \\
If you see only this Streamlit app, you are in the sharing mode, https://share.stlite.net/.
(If you want to edit the app, please go to the [editor mode](https://edit.share.stlite.net/)!)

The app code and data are encoded into the URL as a hash like `https://share.stlite.net/#!ChBz...`,
so you can save, share and restore the app only this the URL.
If you are on the editor page, click the "Open App" link on the top right toolbar to see the standalone app!

You can switch the editor and sharing modes by replacing the host naem in the URL,
`edit.share.stlite.net` and `share.stlite.net`.

### Tell your story!
When you create some apps with _Stlite_, please share it!
All you need to do is copy and paste the URL 👍

* **Stlite** GitHub Discussions [🔗 here](https://github.com/whitphx/stlite/discussions/categories/show-and-tell)
* Streamlit community forum [🔗 here](https://discuss.streamlit.io/)
""")

st.header("Streamlit Component Samples")
st.markdown("""
    All these features are working on your browser!
""")

name = st.text_input("Your name?")
st.write("Hello,", name or "world", "!")

value = st.slider("Value?")
st.write("The slider value is", value)

import numpy as np
import pandas as pd

st.subheader("Chart sample")
chart_data = pd.DataFrame(
    np.random.randn(20, 3),
    columns=['a', 'b', 'c'])

tab1, tab2, tab3 = st.tabs(["Line chart", "Area chart", "Bar chart"])
with tab1:
    st.line_chart(chart_data)
with tab2:
    st.area_chart(chart_data)
with tab3:
    st.bar_chart(chart_data)

st.subheader("DataFrame sample")
df = pd.DataFrame(
   np.random.randn(50, 20),
   columns=('col %d' % i for i in range(20)))

st.dataframe(df)

st.subheader("Camera input")
st.info("Don't worry! The photo data is processed on your browser and never uploaded to any remote servers.")
enable_camera_input = st.checkbox("Use the camera input")
if enable_camera_input:
    picture = st.camera_input("Take a picture")

    if picture:
        st.image(picture)
